// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: AGPL-3.0-or-later

use {
    crate::constants::USER_AGENT,
    anyhow::{anyhow, Result},
    reqwest::{StatusCode, Url},
    serde::{Deserialize, Serialize},
    std::collections::HashMap,
    std::io::ErrorKind,
    std::path::{Path, PathBuf},
    std::time::Duration,
};

#[derive(Deserialize, Serialize, Default)]
pub struct Credentials {
    #[serde(skip)]
    config_file: PathBuf,

    pub servers: HashMap<String, Server>,
}

#[derive(Deserialize, Serialize)]
pub struct Server {
    pub url: String,
    pub login_name: String,
    pub password: String,
}

impl Credentials {
    pub fn from_directory(config_dir: &Path) -> Result<Self> {
        let credentials_file = config_dir.join("credentials.toml");
        match std::fs::read_to_string(&credentials_file) {
            Ok(content) => {
                let mut credentials: Credentials = toml::from_str(&content)?;
                credentials.config_file = credentials_file;
                Ok(credentials)
            }
            Err(err) if err.kind() == ErrorKind::NotFound => {
                let mut credentials = Credentials::default();
                credentials.config_file = credentials_file;
                credentials.write_back()?;
                Ok(credentials)
            }
            Err(err) => Err(anyhow!(err)),
        }
    }

    pub fn add(&mut self, server: Url) -> Result<()> {
        let http_client = reqwest::blocking::Client::builder()
            .https_only(true)
            .user_agent(USER_AGENT)
            .build()?;

        #[derive(Deserialize)]
        struct LoginFlow {
            poll: LoginPollEndpoint,
            login: String,
        }
        #[derive(Deserialize)]
        struct LoginPollEndpoint {
            token: String,
            endpoint: String,
        }
        #[derive(Deserialize)]
        #[serde(rename_all = "camelCase")]
        struct LoginResult {
            server: String,
            login_name: String,
            app_password: String,
        }

        let login_url = server.join("index.php/login/v2")?;

        let login_flow = http_client
            .post(login_url.clone())
            .send()?
            .json::<LoginFlow>()?;

        // User auth flow happens in browser
        webbrowser::open(&login_flow.login)?;

        // We start polling for the end of the flow
        loop {
            let response = http_client
                .post(&login_flow.poll.endpoint)
                .form(&[("token", &login_flow.poll.token)])
                .send()?;

            match response.status() {
                StatusCode::OK => {
                    let login_result = response.json::<LoginResult>()?;
                    let new_server = Server {
                        url: login_result.server,
                        login_name: login_result.login_name,
                        password: login_result.app_password,
                    };

                    let host_name = server
                        .host_str()
                        .ok_or(anyhow!("No hostname for provided URL, is it valid?"))?;
                    self.servers.insert(host_name.to_owned(), new_server);
                    break;
                }
                StatusCode::NOT_FOUND => {
                    std::thread::sleep(Duration::from_secs(1));
                    // ...then keep polling
                }
                _ => {
                    response.error_for_status()?;
                }
            }
        }

        self.write_back()
    }

    fn write_back(&self) -> Result<()> {
        #[cfg(unix)]
        use std::os::unix::fs::PermissionsExt;

        if !self.config_file.exists() {
            let f = std::fs::File::create(&self.config_file)?;
            #[cfg(unix)]
            f.metadata()?.permissions().set_mode(0o600);
        }
        std::fs::write(&self.config_file, toml::to_string(self)?)?;
        Ok(())
    }
}
