// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: AGPL-3.0-or-later

use {
    chrono::Duration,
    serde::{Deserialize, Deserializer},
};

#[derive(Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct Metadata {
    #[serde(rename = "recipe_id")]
    pub id: u64,
    pub name: String,
    pub keywords: Option<String>,

    #[serde(with = "naive_date_format")]
    pub date_created: DateTime,

    #[serde(with = "naive_date_format")]
    pub date_modified: DateTime,
}

/// A recipe according to [schema.org](http://schema.org/Recipe)
#[derive(Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct Recipe {
    pub id: u64,
    pub name: String,
    pub description: String,
    pub url: Option<String>,
    pub keywords: Option<String>,

    #[serde(rename = "dateCreated")]
    pub created: DateTime,

    #[serde(rename = "dateModified")]
    pub modified: Option<DateTime>,

    pub image_url: String,

    #[serde(with = "duration")]
    pub prep_time: Duration,

    #[serde(default)]
    #[serde(with = "maybe_duration")]
    pub cook_time: Option<Duration>,

    #[serde(default)]
    #[serde(with = "maybe_duration")]
    pub total_time: Option<Duration>,

    pub image: Option<String>,
    pub recipe_yield: isize,

    #[serde(rename = "recipeCategory")]
    pub category: Option<String>,

    pub tools: Option<Vec<Tool>>,
    #[serde(rename = "recipeIngredient")]
    pub ingredients: Vec<Ingredient>,
    #[serde(rename = "recipeInstructions")]
    pub instructions: Vec<Instruction>,
    //pub nutrition: Nutrition,
}

#[derive(Debug, PartialEq, PartialOrd)]
pub struct Ingredient {
    pub name: String,
    pub amount: f64,
    pub unit: Unit,
}

#[derive(Debug, Eq, PartialEq, PartialOrd)]
pub enum Unit {
    None,
    Kilogram,
    Liter,
}

#[derive(Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct Tool(pub String);

#[derive(Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct Instruction(pub String);

#[derive(Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct Nutrition {
    pub calories: Option<String>,
    pub carbohydrates_content: Option<String>,
    pub cholesterol_content: Option<String>,
    pub fat_content: Option<String>,
    pub fiber_content: Option<String>,
    pub protein_content: Option<String>,
    pub saturated_fat_content: Option<String>,
    pub serving_size: Option<String>,
    pub sodium_content: Option<String>,
    pub sugar_content: Option<String>,
    pub trans_fat_content: Option<String>,
    pub unsaturated_fat_content: Option<String>,
}

type DateTime = chrono::DateTime<chrono::Utc>;

mod naive_date_format {
    use chrono::{DateTime, TimeZone, Utc};
    use serde::{self, Deserialize, Deserializer};

    const FORMAT: &'static str = "%Y-%m-%d %H:%M:%S";

    pub fn deserialize<'de, D>(deserializer: D) -> Result<DateTime<Utc>, D::Error>
    where
        D: Deserializer<'de>,
    {
        let s = String::deserialize(deserializer)?;
        Utc.datetime_from_str(&s, FORMAT)
            .map_err(serde::de::Error::custom)
    }
}

mod duration {
    use {
        super::Duration,
        serde::{self, de::Error, Deserialize, Deserializer},
    };

    pub fn deserialize<'de, D>(deserializer: D) -> Result<Duration, D::Error>
    where
        D: Deserializer<'de>,
    {
        let value = String::deserialize(deserializer)?;
        parse_duration::<D>(&value)
    }

    pub fn parse_duration<'de, D>(value: &str) -> Result<Duration, D::Error>
    where
        D: Deserializer<'de>,
    {
        speedate::Duration::parse_str(&value)
            .map(|dt| Duration::seconds(dt.signed_total_seconds()))
            .map_err(|e| D::Error::custom(e.to_string()))
    }
}

mod maybe_duration {
    use {
        super::Duration,
        serde::{self, Deserialize, Deserializer},
    };

    pub fn deserialize<'de, D>(deserializer: D) -> Result<Option<Duration>, D::Error>
    where
        D: Deserializer<'de>,
    {
        let s: Option<String> = Option::deserialize(deserializer)?;
        Ok(if let Some(s) = s {
            Some(super::duration::parse_duration::<D>(&s)?)
        } else {
            None
        })
    }
}

impl Recipe {
    pub fn total_time(&self) -> Duration {
        self.total_time
            .unwrap_or_else(|| self.prep_time + self.cook_time.unwrap_or(Duration::zero()))
    }
}

impl<'de> Deserialize<'de> for Ingredient {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        deserializer.deserialize_str(IngredientVisitor)
    }
}

struct IngredientVisitor;

impl<'de> serde::de::Visitor<'de> for IngredientVisitor {
    type Value = Ingredient;

    fn visit_str<E>(self, value: &str) -> Result<Self::Value, E>
    where
        E: serde::de::Error,
    {
        use std::cmp::{max, min};

        let quantity_regex_start: regex::Regex =
            regex::Regex::new(r"^(?P<v>[0-9,\.]+)\s*(?P<u>\w+)?\s+(?P<i>.*)").unwrap();
        let quantity_regex_end: regex::Regex =
            regex::Regex::new(r"(?P<i>.*?)\s+(?P<v>[0-9,\.]+)\s*(?P<u>\w+)?$").unwrap();

        if let Some(captures) = quantity_regex_start
            .captures(value)
            .or_else(|| quantity_regex_end.captures(value))
        {
            let vm = captures.name("v").unwrap();
            let v = vm.as_str();
            let um = captures.name("u");
            let u = um.map(|u| u.as_str()).unwrap_or("");
            let im = captures.name("i").unwrap();
            let i = im.as_str();

            let (mut amount, unit) = Self::parse_quantity(v, u)?;
            let name = match unit {
                Unit::None => {
                    // pick the longest string including what was recognized by the regex as the possible unit
                    let range = um
                        .map(|um| min(um.start(), im.start())..max(um.end(), im.end()))
                        .unwrap_or_else(|| im.start()..im.end());

                    if range.start <= vm.start() && range.end >= vm.end() {
                        amount = 1.0; // e.g. for "Parsley, 5 leaves" -> Ingredient { 1, no unit, 'Parsley, 5 leaves' }
                    }

                    &value[range]
                }
                _ => i,
            }
            .to_owned();

            Ok(Ingredient { name, amount, unit })
        } else {
            Ok(Ingredient {
                name: value.to_owned(),
                amount: 1.0,
                unit: Unit::None,
            })
        }
    }

    fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
        formatter.write_str("a parseable ingredient name with an optional mass or volume quantity at the beginning or end")
    }
}

impl IngredientVisitor {
    fn parse_quantity<E>(value: &str, unit: &str) -> Result<(f64, Unit), E>
    where
        E: serde::de::Error,
    {
        let mut v = value
            .replace(',', ".")
            .parse::<f64>()
            .map_err(|e| serde::de::Error::custom(e.to_string()))?;
        let u = match unit.to_lowercase().as_str() {
            "kg" => Unit::Kilogram,
            "g" | "gr" => {
                v = v / 1000.0;
                Unit::Kilogram
            }
            "hg" | "etto" | "etti" => {
                v = v / 10.0;
                Unit::Kilogram
            }
            "l" => Unit::Liter,
            "dl" => {
                v = v / 10.0;
                Unit::Liter
            }
            "cl" => {
                v = v / 100.0;
                Unit::Liter
            }
            "ml" | "cc" => {
                v = v / 1000.0;
                Unit::Liter
            }
            "tl" | "cucchiaino" | "cucchiaini" => {
                v = v / 1000.0 * 5.0;
                Unit::Liter
            }
            "el" | "cucchiaio" | "cucchiai" => {
                v = v / 1000.0 * 15.0;
                Unit::Liter
            }
            "tsp" => {
                v = v / 1000.0 * 4.93;
                Unit::Liter
            }
            "tblsp" => {
                v = v / 1000.0 * 14.79;
                Unit::Liter
            }
            _ => Unit::None,
        };

        Ok((v, u))
    }
}

impl core::fmt::Display for Ingredient {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        const PRECISION: u32 = 3;
        const ROUNDING_EXP: f64 = 10isize.pow(PRECISION) as f64;

        write!(
            f,
            "{amount}{pad}{unit} {name}",
            name = self.name,
            amount = (self.amount * ROUNDING_EXP).round() / ROUNDING_EXP,
            pad = match self.unit {
                Unit::None => "",
                _ => " ",
            },
            unit = self.unit,
        )
    }
}

impl std::ops::Mul<f64> for Ingredient {
    type Output = Ingredient;

    fn mul(mut self, rhs: f64) -> Self::Output {
        self.amount *= rhs;
        self
    }
}

impl core::fmt::Display for Unit {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let u: &dyn core::fmt::Display = match self {
            Unit::Kilogram => &"kg",
            Unit::Liter => &"L",
            Unit::None => &"",
        };

        write!(f, "{}", u)
    }
}
